import numpy as np
import arithmetics

np.set_printoptions(precision=4, suppress=True)


def cfr(A, B, x0, y0, T, Rx0=0, Ry0=0):
    Rx = x0 * Rx0
    Ry = y0 * Ry0

    for t in range(T):
        x = Rx / Rx.sum(axis=0)
        y = Ry / Ry.sum(axis=0)
        u = A @ y
        v = B.T @ x
        rx = u - u.T @ x
        ry = v - v.T @ y

        Rx += np.maximum(rx, 0)
        Ry += np.maximum(ry, 0)
        # Rx += np.log(1 + np.maximum(rx, 0))
        # Ry += np.log(1 + np.maximum(ry, 0))
        # Rx += np.sin(np.maximum(rx, 0) * 100) ** 2
        # Ry += np.sin(np.maximum(ry, 0) * 100) ** 2
        # Rx += np.random.rand(*rx.shape) * np.maximum(rx, 0) ** 3
        # Ry += np.random.rand(*rx.shape) * np.maximum(ry, 0) ** 3

    x = Rx / Rx.sum(axis=0)
    y = Ry / Ry.sum(axis=0)

    return x, y


def cfr_ep(A, B, x0, y0, T, Rx0=0, Ry0=0):
    Rx = x0 * Rx0
    Ry = y0 * Ry0
    epA = 0
    epB = 0
    ep = 0
    for t in range(T):
        x = Rx / Rx.sum(axis=0)
        y = Ry / Ry.sum(axis=0)
        u = A @ y
        v = B.T @ x
        # rx = u - u.T @ x
        # ry = v - v.T @ y
        epA = arithmetics.smoothmax(A @ y, 0.0001) - x.T @ A @ y
        epB = arithmetics.smoothmax(B.T @ x, 0.0001) - x.T @ B @ y
        ep = epA + epB
        if t % 20 == 0:
            print(ep, x.T, y.T)
        rx = (( A  +  B ) @ y - B @ arithmetics.softmax(B.T @ x, 0.001))
        ry = ((A.T + B.T) @ x - A.T @ arithmetics.softmax(A @ y, 0.001))
        rx -= x.T @ rx
        ry -= y.T @ ry
        Rx += np.maximum(rx, 0)
        Ry += np.maximum(ry, 0)
        Rx *= 0.999
        Ry *= 0.999
    x = Rx / Rx.sum(axis=0)
    y = Ry / Ry.sum(axis=0)

    return x, y, epA, epB, ep


def test():
    A = np.array([[13800, -36365, -18522],
                  [36700, 13800, -36200],
                  [18522, 36365, 13800]]) / 10000
    B = np.array([[29500, 10383, -33139],
                  [21600, 29500, 70700],
                  [33139, -10383, 29500]]) / 10000
    print("A:\n", A)
    print("B:\n", B)

    x0 = arithmetics.rand_f(1, 2).T
    y0 = arithmetics.rand_f(1, 2).T
    x1 = np.array([[0.0000, 0.0690, 0.9310]]).T
    y1 = np.array([[0.7334, 0.0000, 0.2666]]).T

    x, y = cfr(A, B, x0, y0, 2000, Rx0=1, Ry0=1)
    print("cfr:")
    print("x:", x.T)
    print("y:", y.T)
    print("")

    x, y, epA, epB, ep = cfr_ep(A, B, x0, y0, 4, Rx0=20, Ry0=20)
    print("cfr-ep:")
    print(ep, x.T, y.T)
    print("")

    return


if __name__ == "__main__":
    test()
